#!/usr/bin/python

# Copyright (C) 2010, 2011  Evan Broder <evan@ebroder.net>
# Copyright (C) 2010, 2011, 2012  Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import optparse
import re
import sys
import subprocess
import webbrowser

from launchpadlib.errors import ClientError
from launchpadlib.launchpad import Launchpad


CONSUMER_KEY = 'backport-helper'

valid_series = set()
backporters = set()
lp = None


def rmadison(distro, package):
    proc = subprocess.Popen(["/usr/bin/rmadison", "-u", distro, package],
            stdout=subprocess.PIPE)
    (output, _) = proc.communicate()
    assert proc.returncode == 0, "rmadison failed"
    return output.splitlines()


class BPBug(object):
    PACKAGE_RE = re.compile(r'Please backport ([a-z0-9][-a-z0-9+.]+)', re.I)
    SOURCE_RE = re.compile(r'\bfrom ([\w-]+)', re.I)

    def __init__(self, bugtask):
        self.number = bugtask.bug.id
        self.title = bugtask.bug.title
        self.reporter = bugtask.bug.owner
        self.importance = bugtask.importance
        self.status = bugtask.status
        self.assignee = None
        if bugtask.assignee is not None:
            self.assignee = bugtask.assignee.name
        self.milestone = None
        if bugtask.milestone is not None:
            self.milestone = bugtask.milestone.name

        # Try to guess the package
        self.package = None
        self.ubuntu_details = None
        match = self.PACKAGE_RE.search(bugtask.bug.title)
        if match:
            self.setPackage(match.group(1))

        # Try to guess the backport destination
        self.dest = None
        self.setDest(bugtask.bug_target_name)

        # Try to guess the backport source
        self.source = None
        match = self.SOURCE_RE.search(bugtask.bug.title)
        if match:
            self.setSource(match.group(1))

        # Collate the bug's history
        self.content = bugtask.bug.description
        for m in reversed(bugtask.bug.messages):
            # We might get 410 Gone if a message has been deleted as spam.
            try:
                owner = m.owner
                owner.name
            except ClientError:
                continue
            self.content += '\n=============================================\n'
            star = ''
            if owner.name in backporters:
                star = '(*)'
            self.content += '%s%s said %s on %s:\n\n' % (owner.name, star,
                                                         m.subject, str(m.date_created))
            self.content += m.content

        self.requestor = None
        for a in reversed(bugtask.bug.activity):
            if (a.whatchanged == '%s: status' % bugtask.bug_target_name and
                a.newvalue == 'In Progress' and
                a.person.name in backporters):
                self.requestor = a.person
                break

    # All setters return True if the input was valid
    def setPackage(self, value):
        self.package = value
        output = rmadison("ubuntu", self.package)
        new_output = []
        for line in output:
            parts = line.split("|")
            if self.validSuite(parts[2].strip().split('/')[0]):
                new_output.append(line)
        self.ubuntu_details = "\n".join(new_output)
        return True

    def setSource(self, value):
        value = value.lower()
        if self.validSuite(value):
            self.source = value
            return True
        return False

    def setDest(self, value):
        value = value.lower()
        value = value.split('-')[0]
        if self.validSuite(value):
            self.dest = value
            return True
        return False

    def setRequestor(self, value):
        try:
            self.requestor = lp.people[value]
            return True
        except KeyError:
            return False

    def valid(self):
        return self.package and self.source and self.dest and self.requestor

    def validSuite(self, suite):
        return suite.split('-')[0] in valid_series

    def writeBackport(self, f):
        params = {'number': self.number,
                  'fromsuite': self.source,
                  'tosuite': self.dest,
                  'package': self.package,
                  'requestor': self.requestor.name}
        f.write('backport %(number)s -b %(requestor)s -S %(fromsuite)s -s %(tosuite)s %(package)s\n' % params)

    def __str__(self):
        return '#%d (%s: %s -> %s, from %s) %s' % (self.number,
                                                   self.package if self.package else 'unknown package',
                                                   self.source if self.source else 'unknown',
                                                   self.dest if self.dest else 'unknown',
                                                   self.requestor.name if self.requestor else 'unknown',
                                                   self.title)

def length(collection):
    # XXX: Workaround bug 274074. Thanks wgrant.
    return int(collection._wadl_resource.representation['total_size'])


def parse_options(args=None):
    p = optparse.OptionParser(usage='Usage: %prog [options] filename')

    p.add_option('-l', '--launchpad', dest='launchpad_instance',
                 default='production')

    options, args = p.parse_args(args=args)
    if len(args) != 1:
        p.error('Need a filename to write instructions to')

    return options, args


def main(args=None):
    global valid_series, backporters, lp

    options, args = parse_options(args[1:])

    try:
        f = open(args[0], 'w')
    except:
        print('Error opening instruction file')
        return 1

    print('Connecting to LP ... ', end="")
    lp = Launchpad.login_with(CONSUMER_KEY, options.launchpad_instance)
    print('Done')

    valid_series = set(s.name for s in lp.distributions['ubuntu'].series if s.active)
    bp_team = lp.people['ubuntu-backporters']
    backporters = set(bp_team.getMembersByStatus(status='Approved')).union(
        bp_team.getMembersByStatus(status='Administrator'))
    backporters = set(bp.name for bp in backporters)
    pillar = lp.projects['ubp']
    subscriber = lp.people['ubuntu-archive']

    print('Loading bugs (this may take some time) ...')
    bugs = pillar.searchTasks(bug_subscriber=subscriber,
                              status='In Progress')
    bp_bugs = []
    num_bugs = length(bugs)
    print('Found %d bug(s)' % num_bugs)

    for i, bugtask in enumerate(bugs, 1):
        bug = BPBug(bugtask)
        print('%d/%d: %s' % (i, num_bugs, bug))
        bp_bugs.append(bug)
    print('Done, %d bugs to consider.' % len(bp_bugs))

    for bug in bp_bugs:
        print()
        print('==========================')

        full = True
        while True:
            print()
            print(bug)
            print('%s %s' % (bug.status, bug.importance), end="")
            if bug.assignee is not None:
                print(' %s' % bug.assignee, end="")
            if bug.milestone is not None:
                print(' %s' % bug.milestone, end="")
            print()
            if full and bug.ubuntu_details is not None:
                sys.stdout.write(bug.ubuntu_details + "\n")

            msg = '[Show/sKip/'
            if bug.valid():
                msg += 'process Backport/'
            msg += 'set Package/set sOurce series/set Dest series/set Requestor/opEn bug]: '
            sys.stdout.write(msg)
            sys.stdout.flush()
            ret = sys.stdin.readline().rstrip().upper()

            if ret == 'S':
                print(bug.content)
            elif ret == 'K':
                break
            elif ret == 'B':
                if not bug.valid():
                    print('Backport does not yet have all required information')
                    continue
                bug.writeBackport(f)
                break
            elif ret == 'P':
                maybe_package = raw_input('New package? ').lower()
                bug.setPackage(maybe_package)
            elif ret == 'O':
                if not bug.setSource(raw_input('New source series? ')):
                    print('Invalid series')
            elif ret == 'D':
                if not bug.setDest(raw_input('New dest series? ')):
                    print('Invalid series')
            elif ret == 'R':
                if not bug.setRequestor(raw_input('New requestor? ')):
                    print('Invalid requestor')
            elif ret == 'E':
                webbrowser.open('https://bugs.launchpad.net/bugs/%s' % bug.number)
            else:
                print('Unknown command: %s' % ret)
            full = False


if __name__ == '__main__':
    sys.exit(main(sys.argv))
