#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (C) 2008, 2009, 2010, 2011, 2012  Canonical Ltd.
# Copyright (C) 2010  Stéphane Graber <stgraber@stgraber.org>
# Copyright (C) 2010  Michael Bienia <geser@ubuntu.com>
# Copyright (C) 2011  Iain Lane <laney@ubuntu.com>
# Copyright (C) 2011  Soren Hansen <soren@linux2go.dk>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Edit uploader permissions for the Ubuntu distro in Launchpad."""

from __future__ import print_function

import sys
from optparse import OptionParser

import launchpadlib.errors
from launchpadlib.launchpad import Launchpad


CONSUMER_KEY = "edit_acl.py"


def print_perms(perms, series=None):
    for perm in perms:
        if (series is not None and perm.distro_series_name is not None and
            series != perm.distro_series_name):
            continue
        desc = []
        desc.append("archive '%s'" % perm.archive.name)
        if perm.component_name:
            desc.append("component '%s'" % perm.component_name)
            if series:
                desc[-1] += ' in %s' % series
        if perm.package_set_name:
            desc.append("package set '%s' in %s" % (perm.package_set_name,
                                                    perm.distro_series_name))
        if perm.source_package_name:
            desc.append("source package '%s'" % perm.source_package_name)
        print("%s for %s: %s" % (perm.permission, perm.person.name,
                                 ', '.join(desc)))


def get_archive(options, launchpad):
    distros = launchpad.distributions
    ubuntu = distros['ubuntu']
    if options.archive is None:
        return ubuntu.main_archive
    else:
        if '/' in options.archive:
            owner, ppa_name = options.archive.split('/')
            return launchpad.people[owner].getPPAByName(name=ppa_name)
        for archive in ubuntu.archives:
            if archive.name == options.archive:
                return archive
    raise AssertionError("No such archive in Ubuntu: %s" % options.archive)


def get_source_components(options, launchpad, archive, source):
    try:
        from debian import debian_support
    except ImportError:
        from debian_bundle import debian_support

    args = {}
    if options.series:
        args['distro_series'] = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)

    newest = {}
    for spph in archive.getPublishedSources(
            source_name=source, exact_match=True, status='Published', **args):
        if not spph.distro_series.active:
            continue
        new_version = debian_support.Version(spph.source_package_version)
        if (spph.distro_series.name not in newest or
            new_version > newest[spph.distro_series.name][0]):
            newest[spph.distro_series.name] = (new_version,
                                               spph.component_name)

    for series in sorted(newest.keys(), key=lambda s: newest[s][0]):
        yield series, newest[series][1]


permission_names = dict(upload='Archive Upload Rights',
                        admin='Queue Administration Rights')


def do_query(options):
    """Query existing permissions and show on stdout."""
    if options.archive is None:
        archives = launchpad.distributions['ubuntu'].archives
    else:
        archives = [get_archive(options, launchpad)]

    if options.person:
        for person in options.person:
            if '@' in person:
                lp_person = launchpad.people.getByEmail(email=person)
            else:
                try:
                    lp_person = launchpad.people[person]
                except KeyError:
                    print("Person '%s' doesn't exist." % person)
                    sys.exit(1)
            perms = []
            for archive in archives:
                perms.extend(archive.getPermissionsForPerson(
                    person=lp_person))
            if options.acl_type:
                perm_name = permission_names[options.acl_type]
                perms = [p for p in perms if p.permission == perm_name]
            print("== All rights for %s ==" % lp_person.name)
            print_perms(perms, options.series)

    if options.component:
        perms = []
        if not options.acl_type or options.acl_type == 'upload':
            for archive in archives:
                perms.extend(archive.getUploadersForComponent(
                    component_name=options.component))
        if not options.acl_type or options.acl_type == 'admin':
            for archive in archives:
                perms.extend(archive.getQueueAdminsForComponent(
                    component_name=options.component))
        print("== All rights for component '%s' ==" % options.component)
        print_perms(perms, options.series)

    if options.packageset:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)

            perms = []
            for archive in archives:
                perms.extend(archive.getUploadersForPackageset(
                    packageset=lp_set))
            print(("== All uploaders for package set '%s' in '%s' (owned by '%s') ==" %
                  (packageset, options.series, lp_set.owner.display_name)))
            print_perms(perms, options.series)

            sources = sorted(lp_set.getSourcesIncluded(direct_inclusion=True))
            if sources:
                print()
                print("== All source packages in package set '%s' "
                      "in '%s' ==" % (packageset, options.series))
                for source in sources:
                    print(source)
            child_sets = list(lp_set.setsIncluded(direct_inclusion=True))
            if child_sets:
                print()
                print("== All package sets in package set '%s' in '%s' ==" %
                      (packageset, options.series))
                for child_set in child_sets:
                    print(child_set.name)

    if options.source:
        for source in options.source:
            perms = []
            perms_set = []
            for archive in archives:
                perms.extend(archive.getUploadersForPackage(
                    source_package_name=source))
                perms_set.extend(archive.getPackagesetsForSource(
                    sourcepackagename=source))
            print("== All uploaders for package '%s' ==" % source)
            print_perms(perms, options.series)
            print_perms(perms_set, options.series)
            for archive in archives:
                for series, component in get_source_components(
                        options, launchpad, archive, source):
                    perms_component = archive.getUploadersForComponent(
                        component_name=component)
                    print_perms(perms_component, series=series)


def validate_add_delete_options(options):
    if options.packageset and options.source:
        # Special options to manage package sets, bodged into this tool
        # since they aren't entirely inconvenient here.
        if options.component or options.person:
            print("-P <packageset> -s <source> cannot be used with a "
                  "component or person as well")
            return False
        return True

    if not options.person:
        print("You must specify at least one person to authorise.")
        return False

    count = 0
    if options.component:
        count += 1
    if options.packageset:
        count += 1
    if options.source:
        count += 1
    if count > 1:
        print("You can only specify one of package set, source, or component")
        return False

    if count == 0:
        print("You must specify one of package set, source, or component")
        return False

    return True


def do_add(options):
    """Add a new permission."""
    if not validate_add_delete_options(options):
        return False

    archive = get_archive(options, launchpad)

    if options.packageset and options.source:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            lp_set.addSources(names=options.source)
            print("Added:")
            for source in options.source:
                print(source)
        return

    people = [launchpad.people[person] for person in options.person]

    if options.source:
        for source in options.source:
            for person in people:
                perm = archive.newPackageUploader(
                    person=person, source_package_name=source)
                print("Added:")
                print_perms([perm,])
        return

    if options.packageset:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            for person in people:
                perm = archive.newPackagesetUploader(
                    person=person, packageset=lp_set)
                print("Added:")
                print_perms([perm,])
        return

    if options.component:
        for person in people:
            if not options.acl_type or options.acl_type == 'upload':
                perm = archive.newComponentUploader(
                    person=person, component_name=options.component)
            else:
                perm = archive.newQueueAdmin(
                    person=person, component_name=options.component)
            print("Added:")
            print_perms([perm,])
        return


def do_delete(options):
    """Delete a permission."""
    if not validate_add_delete_options(options):
        return False

    archive = get_archive(options, launchpad)

    if options.packageset and options.source:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            lp_set.removeSources(names=options.source)
            print("Deleted:")
            for source in options.source:
                print(source)
        return

    lp_people = [launchpad.people[person] for person in options.person]

    if options.source:
        for source in options.source:
            for lp_person in lp_people:
                try:
                    archive.deletePackageUploader(
                        person=lp_person, source_package_name=source)
                    print("Deleted %s/%s" % (lp_person.name, source))
                except Exception:
                    print("Failed to delete %s/%s" % (lp_person.name, source))
        return

    if options.packageset:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            for lp_person in lp_people:
                archive.deletePackagesetUploader(person=lp_person,
                                                 packageset=lp_set)
                print("Deleted %s/%s/%s" % (lp_person.name, packageset,
                                            options.series))
        return

    if options.component:
        for lp_person in lp_people:
            if not options.acl_type or options.acl_type == 'upload':
                archive.deleteComponentUploader(
                    person=lp_person, component_name=options.component)
                print("Deleted %s/%s" % (lp_person.name, options.component))
            else:
                archive.deleteQueueAdmin(
                    person=lp_person, component_name=options.component)
                print("Deleted %s/%s (admin)" % (lp_person.name,
                                                 options.component))
        return


def do_create(options):
    if not options.packageset:
        print("You can only create a package set, not something else.")
        return False

    if not options.person or len(options.person) != 1:
        print("You must specify exactly one person to own the new package set.")
        return False

    if options.series:
        distro_series = launchpad.distributions['ubuntu'].getSeries(
                name_or_version=options.series)
    else:
        distro_series = launchpad.distributions['ubuntu'].getDevelopmentSeries()[0]

    lp_person = launchpad.people[options.person[0]]

    for packageset in options.packageset:
        try:
            if launchpad.packagesets.getByName(
                    name=packageset, distroseries=distro_series):
                print("Package set %s already exists" % packageset)
                continue
        except (TypeError, launchpadlib.errors.HTTPError):
            pass
        desc = raw_input("Description for new package set %s: " % packageset)
        ps = launchpad.packagesets.new(name=packageset, description=desc,
                                distroseries=distro_series, owner=lp_person)
        print(ps)


def do_copy(options):
    if options.archive is None:
        archives = launchpad.distributions['ubuntu'].archives
    else:
        archives = [get_archive(options, launchpad)]

    if not options.packageset:
        print("You can only copy a package set, not something else.")
        return False

    if options.series:
        distro_series = launchpad.distributions['ubuntu'].getSeries(
                name_or_version=options.series)
    else:
        distro_series = launchpad.distributions['ubuntu'].getDevelopmentSeries()[0]

    dst = raw_input("Name of the destination series: ")
    dst_series = launchpad.distributions['ubuntu'].getSeries(name_or_version=dst)

    for packageset in options.packageset:
        src_pkgset = launchpad.packagesets.getByName(
                        name=packageset, distroseries=distro_series)
        if not src_pkgset:
            print("Package set %s doesn't exist" % packageset)

        ps = launchpad.packagesets.new(name=packageset, description=src_pkgset.description,
                                distroseries=dst_series, owner=src_pkgset.owner_link)
        print(ps)

        ps.addSources(names=src_pkgset.getSourcesIncluded())

        perms = []
        for archive in archives:
            perms.extend(archive.getUploadersForPackageset(
                packageset=src_pkgset))

        for perm in perms:
            archive.newPackagesetUploader(
                person=perm.person_link, packageset=ps)



def do_check(options):
    """Check if a person can upload a package."""
    if not options.person:
        print("A person needs to be specified to check.")
        return False
    if not options.source:
        print("A source package needs to be specified to check.")
        return False

    archive = get_archive(options, launchpad)
    people = [launchpad.people[person] for person in options.person]

    if options.series:
        distro_series = launchpad.distributions['ubuntu'].getSeries(
                name_or_version=options.series)
    else:
        distro_series = launchpad.distributions['ubuntu'].getDevelopmentSeries()[0]

    for person in people:
        for srcpkg in options.source:
            spph = archive.getPublishedSources(
                    distro_series=distro_series,
                    exact_match=True,
                    pocket='Release',
                    source_name=srcpkg,
                    status='Published',
                    )[0]
            try:
                archive.checkUpload(
                        component=spph.component_name,
                        distroseries=distro_series,
                        person=person,
                        pocket='Release',
                        sourcepackagename=srcpkg,
                        )
                print("%s (%s) can upload %s to %s" % (
                        person.display_name, person.name,
                        srcpkg, distro_series.displayname))
            except launchpadlib.errors.HTTPError as e:
                if e.response.status == 403:
                    print("%s (%s) cannot upload %s to %s" % (
                            person.display_name, person.name,
                            srcpkg, distro_series.displayname))
                else:
                    print("There was a %s error:" % e.response.status)
                    print(e.content)


def main(options, action):

    if action == "query":
        do_query(options)
    elif action == "add":
        do_add(options)
    elif action in ("delete", "remove"):
        do_delete(options)
    elif action == "create":
        do_create(options)
    elif action == "copy":
        do_copy(options)
    elif action == "check":
        do_check(options)
    else:
        raise AssertionError("Invalid action %s" % action)

    return


if __name__ == '__main__':
    parser = OptionParser(usage="usage: %prog [options] query|add|delete|create|copy|check")

    parser.add_option(
        "-l", "--launchpad", dest="launchpad_instance", default="production")
    parser.add_option("-A", "--archive", dest="archive")
    parser.add_option("-p", "--person", dest="person", action="append")
    parser.add_option("-c", "--component", dest="component")
    parser.add_option("-P", "--packageset", dest="packageset", action="append")
    parser.add_option("-S", "--series", dest="series")
    parser.add_option("-s", "--source", dest="source", action="append")
    parser.add_option("-t", "--acl-type", dest="acl_type",
                      help="ACL type: upload or admin")
    parser.add_option("--anon", dest="anon_login", action="store_true",
            default=False, help="Login anonymously to Launchpad")

    options, args = parser.parse_args()

    possible_actions = ('query', 'add', 'delete', 'create', 'copy', 'check')

    if len(args) != 1:
        parser.error(
            "You must specify an action, one of:\n%s"
                % ", ".join(possible_actions))

    if options.anon_login and args[0] not in ('query', 'check'):
        print("E: Anonymous login not supported for this action.")
        sys.exit(1)

    if (not options.person and not options.component and
        not options.packageset and not options.source):
        parser.error("Provide at least one of "
                     "person/component/packageset/source")
    if options.packageset and not options.series:
        parser.error("Package set requires an associated series")
    if options.acl_type and options.acl_type not in ('upload', 'admin'):
        parser.error("Invalid ACL type '%s' (valid: 'upload', 'admin')" %
                     options.acl_type)
    if options.acl_type == 'admin' and options.packageset:
        parser.error("ACL type admin not allowed for package sets")
    if options.acl_type == 'admin' and options.source:
        parser.error("ACL type admin not allowed for source packages")

    if options.anon_login:
        launchpad = Launchpad.login_anonymously(
            CONSUMER_KEY, options.launchpad_instance)
    else:
        launchpad = Launchpad.login_with(
            CONSUMER_KEY, options.launchpad_instance)

    try:
        main(options, args[0])
    except launchpadlib.errors.HTTPError as err:
        print("There was a %s error:" % err.response.status)
        print(err.content)
