# -*- coding: utf-8 -*-

# Copyright (C) 2011, 2012  Canonical Ltd.
# Author: Stéphane Graber <stgraber@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# To use this module, you need a ini configuration file at ~/.isotracker.conf
# example:
#  [general]
#  url=http://91.189.93.73/xmlrpc.php
#  username=stgraber
#  password=blablabla
#  default_milestone=Precise Daily

from __future__ import print_function

try:
    from xmlrpc import client as xmlrpclib
except ImportError:
    import xmlrpclib

try:
    import configparser as ConfigParser
except ImportError:
    import ConfigParser

import base64
import os
import sys

# Read configuration
configfile = os.path.expanduser('~/.isotracker.conf')
notefile = os.path.expanduser('~/.isotracker.note')
if not os.path.exists(configfile):
    print("Missing configuration file at: %s" % configfile)
    sys.exit(1)

# Load the config
config = ConfigParser.ConfigParser()
config.read([configfile])

# XMLRPC
url = config.get('general', 'url')
username = config.get('general', 'username')
password = config.get('general', 'password')
try:
    auth = str(base64.b64encode(bytes('%s:%s' % (username, password), 'utf-8')), 'utf-8')
except TypeError:
    auth = base64.b64encode('%s:%s' % (username, password))

class AuthTransport(xmlrpclib.Transport):
    def set_auth(self, auth):
        self.auth = auth

    def get_host_info(self, host):
        host, extra_headers, x509 = \
            xmlrpclib.Transport.get_host_info(self, host)
        if extra_headers is None:
            extra_headers = []
        extra_headers.append(('Authorization', 'Basic %s' % auth))
        return host, extra_headers, x509

transport = AuthTransport()
transport.set_auth(auth)
drupal = xmlrpclib.ServerProxy(url, transport=transport)

# Get some useful data
tracker_products = drupal.qatracker.products.get_list([0])
tracker_milestones = drupal.qatracker.milestones.get_list([0])

def default_milestone():
    try:
        return milestone_code(config.get('general', 'default_milestone'))
    except KeyError:
        raise KeyError("No default milestone selected")

def product_code(product):
    for entry in tracker_products:
        if entry['title'] == product:
            return int(entry['id'])
    else:
        raise KeyError("Product '%s' not found" % product)

def milestone_code(milestone):
    for entry in tracker_milestones:
        if entry['title'] == milestone:
            return int(entry['id'])
    else:
        raise KeyError("Milestone '%s' not found" % milestone)

def get_builds(milestone=None, status=[0, 1]):
    if milestone is None:
        milestone = default_milestone()
    elif not isinstance(milestone, int):
        milestone = milestone_code(milestone)

    return drupal.qatracker.builds.get_list(milestone, status)

def post_build(product, version, milestone=None, note="", notify=True):
    if not isinstance(product, int):
        product = product_code(product)

    if note == "" and os.path.exists(notefile):
        note = open(notefile, 'r').read()

    if milestone is None:
        milestone = default_milestone()
    elif not isinstance(milestone, int):
        milestone = milestone_code(milestone)

    if drupal.qatracker.builds.add(product, milestone, version, note, notify):
        print("Build successfully added to the tracker")
    else:
        print("Failed to add build to the tracker")
