#!/usr/bin/python

# Copyright (C) 2009, 2010, 2011, 2012  Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import os
import subprocess
import sys
import webbrowser

from launchpadlib.launchpad import Launchpad
from launchpadlib.uris import LPNET_SERVICE_ROOT
from launchpadlib.errors import ClientError

APP_NAME = 'sync-helper'
CACHE_DIR = os.path.expanduser('~/.launchpadlib/cache')
SERVICE_ROOT = LPNET_SERVICE_ROOT


class Bug(object):

    def __init__(self, number):
        self.number = number
        self.title = None
        self.package = None
        self.reporter = None
        self.content = None
        self.importance = None
        self.status = None
        self.assignee = None
        self.milestone = None


def length(collection):
    # XXX: Workaround bug 274074. Thanks wgrant.
    return int(collection._wadl_resource.representation['total_size'])

def rmadison(distro, package):
    proc = subprocess.Popen(["rmadison", "-u", distro, package],
            stdout=subprocess.PIPE)
    (output, _) = proc.communicate()
    assert proc.returncode == 0, "rmadison failed"
    return output.splitlines()

debian_series_map = {
    "squeeze": "stable",
    "squeeze-security": "stable-security",
    "wheezy": "testing",
    "sid": "unstable",
}

def get_components(package, debian_details):
    components = {}
    for line in debian_details.splitlines():
        if line.strip() == "":
            continue
        series_component = line.split("|")[2].strip()
        archs = line.split("|")[3].strip().split(", ")
        if "source" in archs:
            if "/" in series_component:
                series, component = series_component.split("/")
            else:
                series = series_component
                component = "main"
            components[series] = component
            if series in debian_series_map:
                components[debian_series_map[series]] = component
    return components

def ubuntu_details(package, current_series):
    output = rmadison("ubuntu", package)
    new_output = []
    for line in output:
        parts = line.split("|")
        if current_series in parts[2]:
            new_output.append(line)
    return new_output

def debian_details(package):
    return rmadison("debian", package)

def canUpload(distribution, person, package_name):
    if distribution.main_archive.isSourceUploadAllowed(
            distroseries=distribution.current_series, person=person,
            sourcepackagename=package_name):
        return True
    pubs = distribution.main_archive.getPublishedSources(
            distro_series=distribution.current_series, exact_match=True,
            pocket="Release", status="Published", source_name=package_name)
    pubs = [pub for pub in pubs]
    if len(pubs) < 1:
        return False
    component = pubs[0].component_name
    for perm in distribution.main_archive.getPermissionsForPerson(person=person):
        if perm.permission != 'Archive Upload Rights':
            continue
        if perm.component_name == component:
            return True
        if perm.source_package_name == package_name:
            return True
    return False

def prompt(bug, suite, full=True):
    sys.stdout.write("%s: %s\n" % (bug.number, bug.title))
    sys.stdout.write("%s %s" % (bug.status, bug.importance))
    if bug.assignee is not None:
        sys.stdout.write(" %s" % (bug.assignee,))
    if bug.milestone is not None:
        sys.stdout.write(" %s" % (bug.milestone,))
    sys.stdout.write("\n")
    if full:
        if bug.ubuntu_details is not None:
            sys.stdout.write(bug.ubuntu_details + "\n")
        if bug.debian_details is not None:
            sys.stdout.write(bug.debian_details + "\n")
    msg = "[Show/sKip/sYnc from %s/" % (suite,)
    if suite == "stable-security":
        msg += "switch to Unstable/"
    else:
        msg += "switch to stAble-security/"
    if suite == "testing":
        msg += "switch to Unstable/"
    else:
        msg += "switch to Testing/"
    if suite == "experimental":
        msg += "switch to Unstable/"
    else:
        msg += "switch to Experimental/"
    msg += "Open bug/change Component]: "
    sys.stdout.write(msg)
    sys.stdout.flush()
    return sys.stdin.readline().rstrip().upper()

def sync(bug, syncfile, suite="unstable", component="main"):
    params = {"bug": bug.number,
              "suite": suite,
              "component": component,
              "name": bug.package,
             }
    if suite.endswith("-security"):
        params["distro"] = "security"
        params["suite"] = "%s/updates" % suite[:-9]
    else:
        params["distro"] = "debian"
    syncfile.write("sync %(bug)s -f -D %(distro)s -S %(suite)s -C %(component)s %(name)s\n" % params)

def main(args):
    f = None
    try:
        f = open(args[1], 'w')
    except:
        print("%s: Require a filename to write to!" % args[0])
        sys.exit(1)
    sys.stdout.write("Connecting to LP ... ")
    sys.stdout.flush()
    launchpad = Launchpad.login_with(APP_NAME, SERVICE_ROOT, CACHE_DIR)
    pillar = launchpad.projects['ubuntu']
    subscriber = launchpad.people['ubuntu-archive']
    ubuntu_series = pillar.current_series.name
    sys.stdout.write("Done\n")
    sys.stdout.write("Loading bugs (this may take some time, go get a "
            "cup of tea) ... \n")
    sync_bugs = []
    bugs = pillar.searchTasks(bug_subscriber=subscriber, search_text='sync')
    num_bugs = length(bugs)
    bugs_processed = 0
    prev_line = ""
    for bugtask in bugs:
        bugs_processed += 1
        sys.stdout.write("\r" + " "*len(prev_line) + "\r")
        prev_line = "%d/%d" % (bugs_processed, num_bugs)
        sys.stdout.write(prev_line)
        sys.stdout.flush()
        if (u"ubuntu" != bugtask.bug_target_name
                and " (Ubuntu)" not in bugtask.bug_target_name):
            continue
        if "debian-multimedia.org" in bugtask.bug.title:
            continue
        bug = Bug(bugtask.bug.id)
        bug.title = "%s: %s" % (bugtask.bug_target_name, bugtask.bug.title)
        bug.package = None
        if " (Ubuntu)" in bugtask.bug_target_name:
            bug.package = bugtask.bug_target_name.rsplit(" ", 1)[0]
            next_line_part = ": %s" % (bug.package,)
            sys.stdout.write(next_line_part)
            prev_line += next_line_part
            sys.stdout.flush()
        else:
            next_line_part = ": Ubuntu"
            sys.stdout.write(next_line_part)
            prev_line += next_line_part
            sys.stdout.flush()
        bug.content = bugtask.bug.description
        bug.reporter = bugtask.bug.owner
        bug.importance = bugtask.importance
        bug.status = bugtask.status
        if bugtask.assignee is not None:
            bug.assignee = bugtask.assignee.name
        if bugtask.milestone is not None:
            bug.milestone = bugtask.milestone.name
        for i in reversed(bugtask.bug.messages):
            # We might get 410 Gone if a message has been deleted as spam.
            try:
                owner = i.owner
                owner.name
            except ClientError:
                continue
            bug.content += "\n=============================================\n"
            star = ""
            if bug.package is not None and canUpload(pillar, owner, bug.package):
                star = "(*)"
            bug.content += "%s%s said %s on %s:\n\n" % (owner.name, star,
                    i.subject, str(i.date_created))
            bug.content += i.content
        if bug.package is not None:
            bug.ubuntu_details = "\n".join(ubuntu_details(bug.package, ubuntu_series))
            bug.debian_details = "\n".join(debian_details(bug.package))
            bug.components = get_components(bug.package, bug.debian_details)
        else:
            bug.ubuntu_details = None
            bug.debian_details = None
            bug.components = None
        sync_bugs.append(bug)
    sys.stdout.write("\r" + " "*len(prev_line) + "\r")
    sys.stdout.write("Done, %d packages to consider.\n" % (len(sync_bugs),))

    for bug in sync_bugs:
        sys.stdout.write("\n==========================\n")
        finished = False
        suite = "unstable"
        if ("from stable-security" in bug.title.lower()
                or "from debian stable-security" in bug.title.lower()):
            suite = "stable-security"
        if ("from testing" in bug.title.lower()
                or "from debian testing" in bug.title.lower()):
            suite = "testing"
        if ("from experimental" in bug.title.lower()
                or "from debian experimental" in bug.title.lower()):
            suite = "experimental"
        component = "main"
        manual_component = False
        if bug.components is not None and suite in bug.components:
            component = bug.components[suite]
        full = True
        while not finished:
            ret = prompt(bug, suite, full=full)
            if ret == 'S':
                print(bug.content)
            elif ret == 'K':
                finished = True
            elif ret == 'Y':
                if bug.package is None:
                    sys.stdout.write("Package to sync: ")
                    sys.stdout.flush()
                    bug.package = sys.stdin.readline().strip()
                sync(bug, f, suite=suite, component=component)
                finished = True
            elif ret == 'A':
                suite = "stable-security"
                if (not manual_component and bug.components is not None
                        and suite in bug.components):
                    component = bug.components[suite]
            elif ret == 'T':
                suite = "testing"
                if (not manual_component and bug.components is not None
                        and suite in bug.components):
                    component = bug.components[suite]
            elif ret == 'U':
                suite = "unstable"
                if (not manual_component and bug.components is not None
                        and suite in bug.components):
                    component = bug.components[suite]
            elif ret == 'E':
                suite = "experimental"
                if (not manual_component and bug.components is not None
                        and suite in bug.components):
                    component = bug.components[suite]
            elif ret == 'O':
                webbrowser.open("https://bugs.launchpad.net/bugs/%s" % bug.number)
            elif ret == 'C':
                sys.stdout.write("Component to sync from: ")
                sys.stdout.flush()
                component = sys.stdin.readline().strip()
                manual_component = True
            else:
                sys.stdout.write("Unknown command: %s\n" % ret)
            full = False
    f.close()
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))

