# Copyright (C) 2009, 2010, 2012  Canonical Ltd.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Some day, we'll have proper APIs for doing everything in Launchpad.  In
# the meantime, we have to shove files through a privileged back door
# (specifically, one that doesn't check signatures, so that we don't have to
# re-GPG-sign all this stuff) to make it work properly.  This is a very
# simple client-side interface to this back door.  It only works if you have
# lp_archive access on the publishing machine.

from __future__ import print_function

import os
import re
import subprocess


if 'UBUNTU_ARCHIVE_HOST' in os.environ:
    remote_host = os.environ['UBUNTU_ARCHIVE_HOST']
else:
    remote_host = 'cocoplum.canonical.com'


non_meta_re = re.compile(r'^[a-zA-Z0-9+,./:=@_-]+$')

# Aggravatingly, ssh runs the command you give it through the shell, so we
# have to shell-escape everything.
def shell_escape(arg):
    if arg:
        if non_meta_re.match(arg):
            return arg
        else:
            return "'%s'" % arg.replace("'", "'\\''")
    return 'None'


def request(op_type, args=None, pipe_stdout=True):
    backend = op_type
    if op_type == 'backport':
        directory = '/home/lp_archive/backports'
        backend = 'backport-source-backend'
    elif op_type == 'flush-backports':
        directory = '/home/lp_archive/backports'
    else:
        directory = '.'

    command = ['ssh', '-q', '-t', remote_host,
               '/home/lp_archive/bin/sync-backend', directory, backend]
    if args is not None:
        command.append(' '.join([shell_escape(arg) for arg in args]))
    if 'SYNCLIB_DEBUG' in os.environ:
        if directory != '.':
            print('cd %s && ' % shell_escape(directory), end='')
        print(backend, command[-1])
        command = ['true']
    if pipe_stdout:
        return subprocess.Popen(command, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT)
    else:
        return subprocess.Popen(command)


def do_sync(op_type, source, extra_opts):
    output = []
    sync_done = True
    opts = list(extra_opts)
    opts.append(source)
    handle = request(op_type, opts)
    for line in handle.stdout:
        print(line, end="")
        if (not line.startswith('Getting binaries for') and
            not line.startswith('No handlers could be found for logger')):
            output.append(line)
        if (line.startswith('[NOT Updating - Modified]') or
            line.startswith('[BLACKLISTED]') or
            line.startswith('[Nothing to update')):
            sync_done = False
    return (handle.wait() == 0 and sync_done, output)


# For syncs, we have a proper API, but it needs some work to do all the
# necessary client-side checks and call it properly; syncpackage from
# ubuntu-dev-tools is good at this.  This would be neater if more of
# syncpackage were in library code, so that we could call it directly rather
# than using a subprocess.
def syncpackage(source, extra_opts, tty=False):
    command = ['syncpackage']
    command.extend(extra_opts)
    command.append(source)
    if 'SYNCLIB_DEBUG' in os.environ:
        print(' '.join(command))
        return (0, '')
    # In principle we could use pexpect to answer syncpackage's questions
    # automatically.  However, that seems like a lot of effort given that
    # the archive administrator interface to this feature is deprecated
    # (i.e. sponsors can simply call syncpackage directly).
    if tty:
        with open("/dev/tty") as tty_file:
            return subprocess.call(command, stdin=tty_file) == 0
    else:
        return subprocess.call(command) == 0
